#!/usr/bin/perl

# bogoprep - preprocessor that helps bogofilter to extract as much information
# from mail as possible.
#
# DON'T USE IT AS A FILTER, OUTPUT OF THIS SCRIPT IS NOT A VALID MESSAGE
#
# This utility is similar to unbase64, although it has more features.
#
# usage: bogoprep.pl [-c] file ...
#
# options:
#   -c - perform conversion
#     Default conversion table converts Polish diacritics encoded in cp-1250
#     and utf-8 to iso-8859-2. You may change that by modifying %conv hash.
#     If you want to do more advanced conversion, consider using eg. konwert.
#
#     FIXME Identifying charset and converting everything to utf8 or ucs16
#     would be certainly the best alternative...
#
# It may be used in .procmailrc as follows:
#
# :0HB
# * ? bogoprep.pl | bogofilter
# { SPAM=yes }
#
# :0:
# * SPAM ?? yes
# IN.spam-catcher
#
# bogoprep:
#   - decodes MIME base64 and quoted-printable encoded attachments (leaving
#     original headers - to prevent loss of information)
#   - decodes base64 and quoted-printable encoded headers (leaving charset
#     and adding special words - "hdrbase64encoded" or "hdrquotedprintablep"
#     may be added, because they may help bogofilter to qualify the message)
#   - removes the content of non-text attachments (as they probably don't
#     contain any useful information for bogofilter)
#   - finds unknown html tags or comments (in text/html attachments) inside
#     words and changes their order (as putting such things is one of methods
#     used by spammers to make lexical analysis harder). For example,
#     "FR<foo>EE !<!-- <bar> baz -->!!" becomes
#     "FREE !!! <foo> <!-- <bar> baz -->". I tried to make html tags
#     recognition as compatible with OE as possible (OE users are probably
#     main target for spammers)
#   - converts \0 to spaces (unbase64 simply deleted them, but it may lead
#     to joining words that should be separate)
#
#     Author:
#     (iso-8859-2) Micha Kosek <michau (at) nowa-huta.krakow.pl>
#     (sorry for my awful English and Perl coding style :>)

use MIME::QuotedPrint;
use MIME::Base64;

#whitelist of html elements that divide words
#TODO check this list
$div = "br|/?(p|div|h[1-6]|table|th|td|tr|textarea)";

%conv = (
"ą" => "", "ć" => "", "ę" => "", "ł" => "", "ń" => "",
"ó" => "", "ś" => "", "ź" => "", "ż" => "",
"Ą" => "", "Ć" => "", "Ę" => "", "Ł" => "", "Ń" => "",
"Ó" => "", "Ś" => "", "Ź" => "", "Ż" => "",
"" => "", "" => "", "" => "", "" => "", "" => "", "" => ""
);

sub convinit
{
  my @c = sort { -(length($a) <=> length($b)) } keys %conv;
  $" = "|";
  $c = "@c";
}

sub convert
{
  my $str = shift;

  $str =~ s/$c/$conv{$&}/ge;
  return $str;
}

sub unbase
{
  my $str = shift;

  $str =~ tr|A-Za-z0-9+/=||cd;
  $str = $lastb64 . $str;
  $len = length($str);
  $rest = $len % 4;
  $lastb64 = substr($str, $len-$rest, 4, "");

  return decode_base64($str);
}

sub output
{
  my $str = shift;

  if(!$html || $h || $h1)
  {
    if ($lasthtml)
    {
      @stragetags = ();
      $htmltext =~ s#<($div)\b# $&#gi;
      sub pushtag { my($pre,$in,$post)=@_; return $in if ($pre =~ /\s/ or $post =~ /\s/); $in =~ s/\0/ /g; push @strangetags, $in; return ""; };
      $htmltext =~ s#(?<=(.))((?:<(?:!(?=--)(?:[^-]|-[^-]|--+[^>-])*-*--|[a-zA-Z!?/](?:[^>]*(?:=\s*(?:"[^"]*"|'[^']*'))+)*[^>]*)>)+)(?=(.))#pushtag($1,$2,$3)#sge;
      $htmltext =~ s/\0/ /g;
      print $htmltext;
      $" = "\n";
      print "@strangetags\n";
      $htmltext = '';
      $lasthtml = 0;
    }
    $str =~ s/\0/ /g;
    print $str;
  }
  else
  {
    $htmltext .= $str;
    $lasthtml = 1;
  }
}

if ($ARGV[0] eq '-c')
{
  $cnv=1;
  shift;
}

convinit if($cnv);
$h=1;
for(<>)
{
  chomp;
  ($h,$h1)=(1,$qp[0]=$qp[1]=$b6[0]=$b6[1]=$txt=$html=0) if(/^From /);
  if($h || $h1)
  {
    if(/^\s/)
    {
      $_="$last$/$_";
    }
    else
    {
      $rfc822=1 if($last=~/^content-type:\s*message\/rfc822\b/si);
      push @bnds, $1 if($last=~/^content-type:\s*multipart\/.*boundary=\s*"([^"]*)"/si);
      $qp[$h1]=1 if ($last=~/^content-transfer-encoding:\s*quoted-printable\s*$/si);
      $b6[$h1]=1 if ($last=~/^content-transfer-encoding:\s*base64\s*$/si);
      $txt=1 if ($last=~/^content-type:\s*text\//si);
      $html=1 if ($last=~/^content-type:\s*text\/html\b/si);
      $last =~ s/=\?([^\?]+)\?B\?(.*?)\?=/"hdrbase64encoded $1 ".decode_base64($2)/geis;
      sub qpd { $x=$_[0];$x=~s/_/=20/gc;return decode_qp($x); }
      $last =~ s/=\?([^\?]+)\?Q\?(.*?)\?=/"hdrquotedprintablep $1 ".qpd($2)." hdrquotedprintableq "/geis;
      $last =~ s/ hdrquotedprintableq\s+hdrquotedprintablep \S+ //gs;
      $last =~ s/ hdrquotedprintableq //gs;
      $last = convert($last) if($cnv);
      output("$last$/") if ($last ne '');
    }
    $last=$_;
  }
  if(/^$/)
  {
    if($rfc822)
    {
      $rfc822=0;
      output("$/");
    }
    else
    {
      $h=$h1=0;
    }
  }
  if(!$h && !$h1)
  {
    if (($bnd=$bnds[@bnds-1]) ne '')
    {
      ($h1,$qp[1],$b6[1],$txt,$html)=(1,0,0,0,0) if($_ eq "--$bnd");
      if($_ eq "--$bnd--")
	{ $h1=$qp[1]=$b6[1]=$txt=$html=0; pop @bnds; }
    }
    if($qp[0]||$qp[1])
    {
      if(/^(.*)=$/)
      {
	$_=$1;
      }
      else
      {
	$_="$_$/";
      }
      $_=decode_qp($_);
      $_=convert($_) if($cnv);
      output($_);
    }
    elsif($b6[0]||$b6[1])
    {
      if($_ eq '')
      {
	output($/);
      }
      elsif($txt)
      {
	$_=unbase($_);
	$_=convert($_) if($cnv);
	s/\r//g;
	output($_);
      }
    }
    else
    {
      $_ = convert($_) if($cnv);
      output("$_$/");
    }
  }
}
$html=0;
output("");
